#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtGui import QStandardItemModel, Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QPushButton, QSplitter, QWidget

from .widgets import NameInput, SelectMenu, SettingsSlider, SettingsTextarea, PresetSelectMenu
from ..utils import trans


class Toolbox:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """
        Setups toolbox

        :return: QSplitter
        """
        # modes and models
        mode_widget = QWidget()
        model_widget = QWidget()
        mode_widget.setLayout(self.setup_list('prompt.mode', trans("toolbox.mode.label")))
        model_widget.setLayout(self.setup_list('prompt.model', trans("toolbox.model.label")))
        mode_widget.setMinimumHeight(150)

        modes_splitter = QSplitter(Qt.Horizontal)
        modes_splitter.addWidget(mode_widget)
        modes_splitter.addWidget(model_widget)

        # presets
        presets = self.setup_presets('preset.presets', trans("toolbox.presets.label"))
        self.window.models['preset.presets'] = self.create_model(self.window)
        self.window.data['preset.presets'].setModel(self.window.models['preset.presets'])

        presets_widget = QWidget()
        presets_widget.setLayout(presets)
        presets_widget.setMinimumHeight(150)

        # initial prompt text
        prompt = self.setup_prompt()

        prompt_widget = QWidget()
        prompt_widget.setLayout(prompt)

        layout = QSplitter(Qt.Vertical)
        layout.addWidget(modes_splitter)  # mode and model list
        layout.addWidget(presets_widget)  # prompts list
        layout.addWidget(prompt_widget)  # prompt text (editable)

        # AI and users names
        names_layout = QHBoxLayout()
        names_layout.addLayout(self.setup_name_input('preset.ai_name', trans("toolbox.name.ai")))
        names_layout.addLayout(self.setup_name_input('preset.user_name', trans("toolbox.name.user")))

        # bottom

        self.window.data['temperature.label'] = QLabel(trans("toolbox.temperature.label"))
        self.window.config_option['current_temperature'] = SettingsSlider(self.window, 'current_temperature',
                                                                          '', 0, 200,
                                                                          1, 100, False)

        self.window.data['img_variants.label'] = QLabel(trans("toolbox.img_variants.label"))
        self.window.config_option['img_variants'] = SettingsSlider(self.window, 'img_variants',
                                                                   '', 1, 4,
                                                                   1, 1, False)

        temp_layout = QVBoxLayout()
        temp_layout.addWidget(self.window.data['temperature.label'])
        temp_layout.addWidget(self.window.config_option['current_temperature'])
        temp_layout.addWidget(self.window.data['img_variants.label'])
        temp_layout.addWidget(self.window.config_option['img_variants'])
        temp_widget = QWidget()
        temp_widget.setLayout(temp_layout)

        bottom_layout = QVBoxLayout()
        bottom_layout.addLayout(names_layout)
        bottom_layout.addWidget(temp_widget)

        names_widget = QWidget()
        names_widget.setLayout(bottom_layout)

        layout.addWidget(names_widget)

        return layout

    def setup_prompt(self):
        """
        Setups preset prompt

        :return: QVBoxLayout
        """
        self.window.data['toolbox.prompt.label'] = QLabel(trans("toolbox.prompt"))
        self.window.data['toolbox.prompt.label'].setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        self.window.data['preset.clear'] = QPushButton(trans('preset.clear'))
        self.window.data['preset.clear'].clicked.connect(
            lambda: self.window.controller.presets.clear())
        self.window.data['preset.use'] = QPushButton(trans('preset.use'))
        self.window.data['preset.use'].clicked.connect(
            lambda: self.window.controller.presets.use())
        self.window.data['preset.use'].setVisible(False)
        header = QHBoxLayout()
        header.addWidget(self.window.data['toolbox.prompt.label'])
        header.addWidget(self.window.data['preset.use'], alignment=Qt.AlignRight)
        header.addWidget(self.window.data['preset.clear'], alignment=Qt.AlignRight)

        self.window.data['preset.prompt'] = SettingsTextarea(self.window, 'preset.prompt', True)
        layout = QVBoxLayout()
        layout.addLayout(header)
        layout.addWidget(self.window.data['preset.prompt'])
        return layout

    def setup_list(self, id, title):
        """
        Setups list

        :param id: ID of the list
        :param title: Title of the list
        :return: QVBoxLayout
        """
        label_key = id + '.label'
        self.window.data[label_key] = QLabel(title)
        self.window.data[label_key].setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        self.window.data[id] = SelectMenu(self.window, id)
        layout = QVBoxLayout()
        layout.addWidget(self.window.data[label_key])
        layout.addWidget(self.window.data[id])

        self.window.models[id] = self.create_model(self.window)
        self.window.data[id].setModel(self.window.models[id])
        return layout

    def setup_presets(self, id, title):
        """
        Setups list

        :param id: ID of the list
        :param title: Title of the list
        :return: QVBoxLayout
        """
        self.window.data['preset.presets.new'] = QPushButton(trans('preset.new'))
        self.window.data['preset.presets.new'].clicked.connect(
            lambda: self.window.controller.presets.edit())

        self.window.data['preset.presets.label'] = QLabel(title)
        self.window.data['preset.presets.label'].setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        header = QHBoxLayout()
        header.addWidget(self.window.data['preset.presets.label'])
        header.addWidget(self.window.data['preset.presets.new'], alignment=Qt.AlignRight)
        header.setContentsMargins(0, 0, 0, 0)

        header_widget = QWidget()
        header_widget.setLayout(header)

        self.window.data[id] = PresetSelectMenu(self.window, id)
        layout = QVBoxLayout()
        layout.addWidget(header_widget)
        layout.addWidget(self.window.data[id])

        self.window.models[id] = self.create_model(self.window)
        self.window.data[id].setModel(self.window.models[id])
        return layout

    def setup_name_input(self, id, title):
        """
        Setups name input

        :param id: ID of the input
        :param title: Title of the input
        :return: QVBoxLayout
        """
        label_key = 'toolbox.' + id + '.label'
        self.window.data[label_key] = QLabel(title)
        self.window.data[id] = NameInput(self.window, id)
        layout = QVBoxLayout()
        layout.addWidget(self.window.data[label_key])
        layout.addWidget(self.window.data[id])
        return layout

    def create_model(self, parent):
        """
        Creates list model
        :param parent: parent widget
        :return: QStandardItemModel
        """
        model = QStandardItemModel(0, 1, parent)
        return model

    def update_list(self, id, data):
        """
        Updates list

        :param id: ID of the list
        :param data: Data to update
        """
        self.window.models[id].removeRows(0, self.window.models[id].rowCount())
        i = 0
        for n in data:
            if 'name' in data[n]:
                self.window.models[id].insertRow(i)
                name = data[n]['name']
                if id == 'preset.presets':
                    if not n.startswith('current.'):
                        name = data[n]['name'] + ' (' + str(n) + ')'
                self.window.models[id].setData(self.window.models[id].index(i, 0), name)
                i += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel
from .debug.config import ConfigDebug
from .debug.context import ContextDebug
from .debug.presets import PresetsDebug
from .debug.models import ModelsDebug


class Debug:
    DBG_KEY, DBG_VALUE = range(2)

    def __init__(self, window=None):
        """
        Debug handler

        :param window: main UI window object
        """
        self.window = window

        # setup workers
        self.workers = {}
        self.workers['config'] = ConfigDebug(self.window)
        self.workers['context'] = ContextDebug(self.window)
        self.workers['presets'] = PresetsDebug(self.window)
        self.workers['models'] = ModelsDebug(self.window)

        # prepare debug ids
        self.ids = self.workers.keys()
        self.models = {}
        self.initialized = {}
        self.active = {}
        self.idx = {}
        self.counters = {}

        # prepare debug workers data
        for id in self.ids:
            self.models[id] = self.create_model(self.window)
            self.initialized[id] = False
            self.active[id] = False
            self.idx[id] = 0

    def update(self, all=False):
        """
        Updates debug windows

        :param all: update all debug windows
        """
        not_realtime = ['context']
        for id in self.workers:
            if id in self.active and self.active[id]:
                if all or id not in not_realtime:
                    self.workers[id].update()

    def begin(self, id):
        """
        Begins debug data

        :param id: debug id
        """
        self.window.debug[id].setModel(self.models[id])
        if id not in self.counters or self.counters[id] != self.models[id].rowCount():
            self.models[id].removeRows(0, self.models[id].rowCount())
            self.initialized[id] = False
        self.idx[id] = 0

    def end(self, id):
        """
        Ends debug data

        :param id: debug id
        """
        self.counters[id] = self.idx[id]
        self.initialized[id] = True

    def add(self, id, k, v):
        """
        Appends debug entry

        :param id: debug id
        :param k: key
        :param v: value
        """
        if self.initialized[id] is False:
            idx = self.models[id].rowCount()
            self.models[id].insertRow(idx)
            self.models[id].setData(self.models[id].index(idx, self.DBG_KEY), k)
            self.models[id].setData(self.models[id].index(idx, self.DBG_VALUE), v)
        else:
            for idx in range(0, self.models[id].rowCount()):
                if self.models[id].index(idx, self.DBG_KEY).data() == k:
                    self.models[id].setData(self.models[id].index(idx, self.DBG_VALUE), v)
                    self.idx[id] += 1
                    return
        self.idx[id] += 1

    def create_model(self, parent):
        """
        Creates list model

        :param parent: parent widget
        :return: model instance
        """
        model = QStandardItemModel(0, 2, parent)
        model.setHeaderData(self.DBG_KEY, Qt.Horizontal, "Key")
        model.setHeaderData(self.DBG_VALUE, Qt.Horizontal, "Value")
        return model



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from ..tokens import num_tokens_prompt, num_tokens_extra
from ..utils import trans


class UI:
    def __init__(self, window=None):
        """
        UI update controller

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """Setups UI"""
        self.update()

    def update(self):
        """Updates UI"""
        self.update_counters()

    def update_counters(self):
        """Updates tokens counters"""
        model = self.window.config.data['model']
        user_name = self.window.config.data['user_name']
        max_total_tokens = self.window.config.data['max_total_tokens']
        extra_tokens = num_tokens_extra(model)

        # prompt tokens
        prompt_tokens = num_tokens_prompt(self.window.config.data['prompt'], user_name, model)

        # input tokens
        input_text = self.window.data['input'].toPlainText().strip()
        input_tokens = num_tokens_prompt(input_text, user_name, model)

        # used tokens
        used_tokens = prompt_tokens + input_tokens

        # context tokens
        ctx_len_all = len(self.window.gpt.context.items)
        ctx_len, ctx_tokens = self.window.gpt.context.count_prompt_items(model, used_tokens, max_total_tokens)

        # zero if context not used
        if not self.window.config.data['use_context']:
            ctx_tokens = 0
            ctx_len = 0

        # total tokens
        total_tokens = prompt_tokens + input_tokens + ctx_tokens + extra_tokens

        # update counters
        string = "{} / {} ~ {} {}".format(ctx_len, ctx_len_all, ctx_tokens, trans('ctx.tokens'))
        self.window.data['prompt.context'].setText(string)

        threshold = str(int(self.window.config.data['context_threshold']))

        string = "~ {} + {} + {} + {} = {} / {} (-{})".format(input_tokens, prompt_tokens, ctx_tokens, extra_tokens,
                                                              total_tokens,
                                                              str(int(max_total_tokens)), threshold)
        self.window.data['input.counter'].setText(string)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import os


class Settings:
    def __init__(self, window=None):
        """
        Settings handler

        :param window: main window
        """
        self.window = window

        # prepare cfg ids
        self.ids = ['settings', 'editor']
        self.active = {}

        # prepare active
        for id in self.ids:
            self.active[id] = False

    def load_default_settings(self):
        """Loads defaults"""
        self.window.config.load_config()
        self.window.controller.settings.init('settings')
        self.window.ui.dialogs.alert("Loaded defaults")

    def load_default_editor(self):
        """Loads defaults from file"""
        file = self.window.dialog['config.editor'].file
        self.load_editor(file)
        self.window.set_status("Loaded defaults from file: {}".format(file))

    def save_editor(self):
        """Saves file to disk"""
        file = self.window.dialog['config.editor'].file
        path = os.path.join(self.window.config.path, file)
        try:
            with open(path, 'w') as f:
                f.write(self.window.editor['config'].toPlainText())
                f.close()
            self.window.set_status("Saved file: {}".format(path))
            self.window.ui.dialogs.alert("Saved file: {}".format(path))
        except Exception as e:
            self.window.set_status("Error saving file: {}".format(path))
            print(e)

    def load_editor(self, file=None):
        """
        Loads file to editor

        :param id: file id
        """
        # load file
        path = os.path.join(self.window.config.path, file)
        self.window.path_label['config'].setText(path)
        self.window.dialog['config.editor'].file = file
        try:
            with open(path, 'r') as f:
                txt = f.read()
                f.close()
                self.window.editor['config'].setPlainText(txt)
        except Exception as e:
            print(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.12 06:00:00                  #
# ================================================== #

import sys

from PySide6.QtGui import QScreen
from PySide6.QtCore import QTimer
from PySide6.QtWidgets import (QApplication, QMainWindow)
from qt_material import QtStyleTools

from .config import Config
from .ui.main import UI
from .controller.main import Controller
from .debugger import Debug
from .settings import Settings
from .info import Info
from .gpt import Gpt
from .image import Image
from .utils import get_init_value


class MainWindow(QMainWindow, QtStyleTools):
    def __init__(self):
        """App main window"""
        super().__init__()
        self.timer = None
        self.github = get_init_value("__github__")
        self.website = get_init_value("__website__")
        self.version = get_init_value("__version__")
        self.build = get_init_value("__build__")
        self.author = get_init_value("__author__")
        self.email = get_init_value("__email__")
        self.data = {}

        # setup config
        self.config = Config()
        self.config.init()

        # app controller
        self.controller = Controller(self)
        self.debugger = Debug(self)
        self.info = Info(self)
        self.settings = Settings(self)

        # handle config migration if needed
        self.controller.launcher.migrate_version()

        # setup GPT and DALL-E
        self.gpt = Gpt(self.config)
        self.images = Image(self.config)

        # setup UI
        self.ui = UI(self)
        self.ui.setup()

        self.setup()
        self.setWindowTitle('PYGPT.net v{} | build {}'.format(self.version, self.build))

    def set_theme(self, theme='dark_teal.xml'):
        """
        Updates theme

        :param theme: theme name
        """
        label = "#ffffff"
        inverse = False
        if theme.startswith('light_'):
            label = "#000000"
            inverse = True
        extra = {
            'density_scale': '-2',
            # 'linux': True,
            'pyside6': True,
            'QLineEdit': {
                'color': label,
            },
            # 'font_family': 'Roboto',
        }
        self.setStyleSheet(self.controller.theme.get_style('line_edit'))
        self.apply_stylesheet(self, theme, invert_secondary=inverse, extra=extra)

    def setup(self):
        """Setups app"""
        self.controller.setup()
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(30)

    def update(self):
        """Called on update"""
        self.debugger.update()

    def set_status(self, text):
        """
        Updates status text

        :param text: status text
        """
        self.data['status'].setText(str(text))

    def closeEvent(self, event):
        """
        Handles close event

        :param event: close event
        """
        print("Closing...")
        print("Saving config...")
        self.config.save_config()
        print("Saving presets...")
        self.config.save_presets()
        print("Exiting...")
        event.accept()  # let the window close


def except_hook(cls, exception, traceback):
    """
    Temporally hook for exceptions handling

    :param cls: class
    :param exception: exception
    :param traceback: traceback
    """
    sys.__excepthook__(cls, exception, traceback)


def run():
    """Runs app"""
    # sys.excepthook = except_hook
    app = QApplication(sys.argv)
    window = MainWindow()
    available_geometry = window.screen().availableGeometry()
    topLeftPoint = QScreen.availableGeometry(QApplication.primaryScreen()).topLeft()
    window.resize(available_geometry.width(), available_geometry.height())
    window.showMaximized()
    window.move(topLeftPoint)

    try:
        sys.exit(app.exec())
    except SystemExit:
        print("Closing...")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

class Confirm:
    def __init__(self, window=None):
        """
        Confirm dialog controller

        :param window: main window object
        """
        self.window = window

    def accept(self, type, id):
        """
        Confirm dialog accept

        :param type: dialog type
        :param id: dialog object id
        """
        if type == 'preset_exists':
            self.window.controller.presets.save(True)
        elif type == 'preset_delete':
            self.window.controller.presets.delete(id, True)
        elif type == 'preset_clear':
            self.window.controller.presets.clear(True)
        elif type == 'ctx_delete':
            self.window.controller.context.delete(id, True)
        elif type == 'ctx_delete_all':
            self.window.controller.context.delete_history(True)
        elif type == 'img_delete':
            self.window.controller.image.img_action_delete(id, True)

        self.window.dialog['confirm'].close()

    def dismiss(self, type, id):
        """
        Confirm dialog dismiss

        :param type: dialog type
        :param id: dialog object id
        """
        self.window.dialog['confirm'].close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from ..utils import trans


class Model:
    def __init__(self, window=None):
        """
        Model and preset select controller

        :param window: main window object
        """
        self.window = window

    def setup(self):
        """Setups all select lists"""
        self.update()

    def select(self, id, value):
        """
        Selects mode, model or preset

        :param id: ID of the list
        :param value: value of the list (row index)
        """
        if id == 'prompt.mode':
            mode = self.window.config.get_mode_by_idx(value)
            self.window.config.data['mode'] = mode
            self.window.config.data['model'] = ""
            self.window.config.data['preset'] = ""
        elif id == 'prompt.model':
            mode = self.window.config.data['mode']
            model = self.window.config.get_model_by_idx(value, mode)
            self.window.config.data['model'] = model
            self.window.config.data['current_model'][mode] = model
        elif id == 'preset.presets':
            mode = self.window.config.data['mode']
            preset = self.window.config.get_preset_by_idx(value, mode)
            self.window.config.data['preset'] = preset
            self.window.config.data['current_preset'][mode] = preset

        self.update()

    def reset_preset_data(self):
        """Resets preset data"""
        self.window.data['preset.prompt'].setPlainText("")
        self.window.data['preset.ai_name'].setText("")
        self.window.data['preset.user_name'].setText("")

    def reset_current_data(self):
        """Resets current data"""
        self.window.config.data['prompt'] = None
        self.window.config.data['ai_name'] = None
        self.window.config.data['user_name'] = None

    def update_preset_data(self):
        """Updates preset data"""
        preset = self.window.config.data['preset']
        if preset is None or preset == "":
            self.reset_preset_data()  # clear preset fields
            self.reset_current_data()
            return

        if preset not in self.window.config.presets:
            self.window.config.data['preset'] = ""  # clear preset if not found
            self.reset_preset_data()  # clear preset fields
            self.reset_current_data()
            return

        # update preset fields
        preset_data = self.window.config.presets[preset]
        self.window.data['preset.prompt'].setPlainText(preset_data['prompt'])
        self.window.data['preset.ai_name'].setText(preset_data['ai_name'])
        self.window.data['preset.user_name'].setText(preset_data['user_name'])

        # update current data
        self.window.config.data['prompt'] = preset_data['prompt']
        self.window.config.data['ai_name'] = preset_data['ai_name']
        self.window.config.data['user_name'] = preset_data['user_name']

    def select_mode_by_current(self):
        """Selects mode by current"""
        mode = self.window.config.data['mode']
        items = self.window.config.get_modes()
        idx = list(items.keys()).index(mode)
        current = self.window.models['prompt.mode'].index(idx, 0)
        self.window.data['prompt.mode'].setCurrentIndex(current)

    def select_model_by_current(self):
        """Selects model by current"""
        mode = self.window.config.data['mode']
        model = self.window.config.data['model']
        items = self.window.config.get_models(mode)
        if model in items:
            idx = list(items.keys()).index(model)
            current = self.window.models['prompt.model'].index(idx, 0)
            self.window.data['prompt.model'].setCurrentIndex(current)

    def select_preset_by_current(self):
        """Selects preset by current"""
        mode = self.window.config.data['mode']
        preset = self.window.config.data['preset']
        items = self.window.config.get_presets(mode)
        if preset in items:
            idx = list(items.keys()).index(preset)
            current = self.window.models['preset.presets'].index(idx, 0)
            self.window.data['preset.presets'].setCurrentIndex(current)

    def update_list_modes(self):
        """Updates modes list"""
        # update modes
        items = self.window.config.get_modes()
        self.window.ui.toolbox.update_list('prompt.mode', items)

    def update_list_models(self):
        """Updates models list"""
        # update modes
        mode = self.window.config.data['mode']
        items = self.window.config.get_models(mode)
        self.window.ui.toolbox.update_list('prompt.model', items)

    def update_list_presets(self):
        """Updates presets list"""
        # update model
        mode = self.window.config.data['mode']
        items = self.window.config.get_presets(mode)
        self.window.ui.toolbox.update_list('preset.presets', items)

    def update_chat_label(self):
        """Updates chat label"""
        mode = self.window.config.data['mode']
        model = self.window.config.data['model']
        if model is None or model == "":
            model_str = "{}".format(trans("mode." + mode))
        else:
            model_str = "{} ({})".format(trans("mode." + mode), model)
        self.window.data['chat.model'].setText(model_str)

    def select_default(self):
        """Sets default mode, model and preset"""
        if self.window.config.data['mode'] is None or self.window.config.data['mode'] == "":
            self.window.config.data['mode'] = self.window.config.get_default_mode()

        if self.window.config.data['model'] is None or self.window.config.data['model'] == "":
            mode = self.window.config.data['mode']

            # set previous selected model
            if mode in self.window.config.data['current_model'] and \
                    self.window.config.data['current_model'][mode] is not None and \
                    self.window.config.data['current_model'][mode] != "" and \
                    self.window.config.data['current_model'][mode] in self.window.config.get_models(mode):
                self.window.config.data['model'] = self.window.config.data['current_model'][mode]
            else:
                # or set default model
                self.window.config.data['model'] = self.window.config.get_default_model(self.window.config.data['mode'])

        if self.window.config.data['preset'] is None or self.window.config.data['preset'] == "":
            mode = self.window.config.data['mode']

            # set previous selected preset
            if mode in self.window.config.data['current_preset'] and \
                    self.window.config.data['current_preset'][mode] is not None and \
                    self.window.config.data['current_preset'][mode] != "" and \
                    self.window.config.data['current_preset'][mode] in self.window.config.get_presets(mode):
                self.window.config.data['preset'] = self.window.config.data['current_preset'][mode]
            else:
                # or set default preset
                self.window.config.data['preset'] = self.window.config.get_default_preset(
                    self.window.config.data['mode'])

    def update_current_temperature(self, temperature=None):
        """
        Updates current temperature

        :param temperature: temperature (float)
        """
        if temperature is None:
            if self.window.config.data['preset'] is None or self.window.config.data['preset'] == "":
                temperature = 1.0  # default temperature
            else:
                preset = self.window.config.data['preset']
                if preset in self.window.config.presets and 'temperature' in self.window.config.presets[preset]:
                    temperature = float(self.window.config.presets[preset]['temperature'])
        self.window.controller.settings.apply("current_temperature", temperature)

    def update_current(self):
        """Updates current mode, model and preset"""
        mode = self.window.config.data['mode']
        if self.window.config.data['preset'] is not None and self.window.config.data['preset'] != "":
            if self.window.config.data['preset'] in self.window.config.presets:
                preset = self.window.config.presets[self.window.config.data['preset']]
                self.window.config.data['user_name'] = preset['user_name']
                self.window.config.data['ai_name'] = preset['ai_name']
                self.window.config.data['prompt'] = preset['prompt']
                self.window.config.data['temperature'] = preset['temperature']
                return

        self.window.config.data['user_name'] = None
        self.window.config.data['ai_name'] = None
        self.window.config.data['temperature'] = 1.0

        # set default prompt if mode is chat
        if mode == 'chat':
            self.window.config.data['prompt'] = self.window.config.data['default_prompt']
        else:
            self.window.config.data['prompt'] = None

    def update_active(self):
        """Updates active mode, model and preset"""
        mode = self.window.config.data['mode']
        if mode == 'chat':
            self.window.data['preset.ai_name'].setDisabled(True)
            self.window.data['preset.user_name'].setDisabled(True)
            self.window.config_option['current_temperature'].slider.setDisabled(False)
            self.window.config_option['current_temperature'].input.setDisabled(False)
            self.window.data['preset.clear'].setVisible(True)
            self.window.data['preset.use'].setVisible(False)

            self.window.data['img_variants.label'].setVisible(False)
            self.window.config_option['img_variants'].setVisible(False)
            self.window.data['temperature.label'].setVisible(True)
            self.window.config_option['current_temperature'].setVisible(True)
        elif mode == 'img':
            self.window.data['preset.prompt'].setDisabled(False)
            self.window.data['preset.ai_name'].setDisabled(True)
            self.window.data['preset.user_name'].setDisabled(True)
            self.window.config_option['current_temperature'].slider.setDisabled(True)
            self.window.config_option['current_temperature'].input.setDisabled(True)
            self.window.data['preset.clear'].setVisible(False)
            self.window.data['preset.use'].setVisible(True)

            self.window.data['img_variants.label'].setVisible(True)
            self.window.config_option['img_variants'].setVisible(True)
            self.window.data['temperature.label'].setVisible(False)
            self.window.config_option['current_temperature'].setVisible(False)
        elif mode == 'completion':
            self.window.data['preset.prompt'].setDisabled(False)
            self.window.data['preset.ai_name'].setDisabled(False)
            self.window.data['preset.user_name'].setDisabled(False)
            self.window.config_option['current_temperature'].slider.setDisabled(False)
            self.window.config_option['current_temperature'].input.setDisabled(False)
            self.window.data['preset.clear'].setVisible(True)
            self.window.data['preset.use'].setVisible(False)

            self.window.data['img_variants.label'].setVisible(False)
            self.window.config_option['img_variants'].setVisible(False)
            self.window.data['temperature.label'].setVisible(True)
            self.window.config_option['current_temperature'].setVisible(True)

    def update(self):
        """Updates all lists"""
        self.select_default()  # set default mode, model and preset if not set
        self.update_current()  # update current mode, model and preset

        # update preset data and current temperature
        self.update_preset_data()
        self.update_current_temperature()

        # update lists
        self.update_list_modes()
        self.update_list_models()
        self.update_list_presets()

        # select current mode, model and preset
        self.select_mode_by_current()
        self.select_model_by_current()
        self.select_preset_by_current()

        # update chat label
        self.update_chat_label()

        # disable / enable widgets
        self.update_active()

        # update tokens counters, etc.
        self.window.controller.ui.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QCheckBox, QWidget, QSplitter

from ..ui.widgets import ChatOutput
from ..ui.input import Input
from ..utils import trans


class ChatBox:
    def __init__(self, window=None):
        """
        Chatbox UI

        :param window: main window object
        """
        self.window = window
        self.input = Input(window)

    def setup(self):
        """
        Setups chatbox

        :return: QSplitter
        """
        self.window.layout_input = self.input.setup()

        self.window.data['output'] = ChatOutput(self.window)
        self.window.data['chat.model'] = QLabel("")
        self.window.data['chat.model'].setAlignment(Qt.AlignRight)
        context_layout = self.setup_context()

        self.window.data['chat.label'] = QLabel(trans("chatbox.label"))
        self.window.data['chat.label'].setStyleSheet(self.window.controller.theme.get_style('text_bold'))

        header = QHBoxLayout()
        header.addWidget(self.window.data['chat.label'])
        header.addWidget(self.window.data['chat.model'])

        layout = QVBoxLayout()
        layout.addLayout(header)
        layout.addWidget(self.window.data['output'])
        layout.addLayout(context_layout)

        output_widget = QWidget()
        output_widget.setLayout(layout)

        input_widget = QWidget()
        input_widget.setLayout(self.window.layout_input)

        # main vertical splitter
        vsplitter = QSplitter(Qt.Vertical)
        vsplitter.addWidget(output_widget)
        vsplitter.addWidget(input_widget)
        vsplitter.setStretchFactor(0, 4)
        vsplitter.setStretchFactor(1, 1)

        return vsplitter

    def setup_context(self):
        """
        Setups context

        :return: QHBoxLayout        
        """
        self.window.data['output.timestamp'] = QCheckBox(trans('output.timestamp'))
        self.window.data['output.timestamp'].setStyleSheet(
            self.window.controller.theme.get_style('checkbox'))  # Windows style fix
        self.window.data['output.timestamp'].stateChanged.connect(
            lambda: self.window.controller.output.toggle_timestamp(self.window.data['output.timestamp'].isChecked()))

        self.window.data['prompt.context'] = QLabel("")
        self.window.data['prompt.context'].setAlignment(Qt.AlignRight)

        layout = QHBoxLayout()
        layout.addWidget(self.window.data['output.timestamp'])
        layout.addWidget(self.window.data['prompt.context'])

        return layout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from .dialog.settings import Settings
from .dialog.preset import Preset
from .dialog.debug import Debug
from .dialog.about import About
from .dialog.changelog import Changelog
from .dialog.editor import Editor
from .dialog.ctx_rename import CtxRename
from .dialog.start import Start
from .dialog.update import Update
from .dialog.image import Image
from .widgets import AlertDialog, ConfirmDialog


class Dialogs:
    def __init__(self, window=None):
        """
        Dialogs setup

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """Setups dialogs"""
        self.window.dialog = {}
        self.window.debug = {}
        self.window.editor = {}

        # setup debug dialogs
        debug = Debug(self.window)
        for id in self.window.debugger.ids:
            debug.setup(id)

        # setup info dialogs
        about = About(self.window)
        changelog = Changelog(self.window)
        about.setup()
        changelog.setup()

        # setup settings dialog
        settings = Settings(self.window)
        settings.setup()

        # setup preset editor dialog
        preset = Preset(self.window)
        preset.setup()

        # setup editor dialog
        editor = Editor(self.window)
        editor.setup()

        # setup ctx rename dialog
        ctx_rename = CtxRename(self.window)
        ctx_rename.setup()

        # setup start dialog
        start = Start(self.window)
        start.setup()

        # setup update dialog
        update = Update(self.window)
        update.setup()

        # setup image dialog
        image = Image(self.window)
        image.setup()

        # alert / confirm
        self.window.dialog['alert'] = AlertDialog(self.window)
        self.window.dialog['confirm'] = ConfirmDialog(self.window)

    def confirm(self, type, id, msg):
        """
        Shows confirm dialog

        :param type: confirm type
        :param id: confirm object id
        :param msg: message to show
        """
        self.window.dialog['confirm'].type = type
        self.window.dialog['confirm'].id = id
        self.window.dialog['confirm'].message.setText(msg)
        self.window.dialog['confirm'].show()

    def alert(self, msg):
        """
        Shows alert dialog

        :param msg: message to show
        """
        self.window.dialog['alert'].message.setPlainText(msg)
        self.window.dialog['alert'].show()

    def open_editor(self, id, data_id, width=400, height=400):
        """
        Opens editor dialog

        :param id: debug dialog id
        :param data_id: data id
        :param width: dialog width
        :param height: dialog height
        """
        if id not in self.window.dialog:
            return
        self.window.dialog[id].resize(width, height)
        self.window.dialog[id].data_id = data_id
        self.window.dialog[id].show()

    def open(self, id, width=400, height=400):
        """
        Opens debug/config dialog

        :param id: debug dialog id
        :param width: dialog width
        :param height: dialog height
        """
        if id not in self.window.dialog:
            return
        self.window.dialog[id].resize(width, height)
        qr = self.window.dialog[id].frameGeometry()
        cp = self.window.screen().availableGeometry().center()
        qr.moveCenter(cp)
        self.window.dialog[id].move(qr.topLeft())
        self.window.dialog[id].show()

    def close(self, id):
        """
        Closes debug/config dialog

        :param id: debug dialog id
        """
        if id not in self.window.dialog:
            return
        self.window.dialog[id].close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

class Debug:
    def __init__(self, window=None):
        """
        Debug controller

        :param window: main window
        """
        self.window = window

    def toggle(self, id):
        """
        Toggles debug window

        :param id: window to toggle
        """
        if id in self.window.debugger.active and self.window.debugger.active[id]:
            self.window.ui.dialogs.close('debug.' + id)
            self.window.debugger.active[id] = False
        else:
            self.window.ui.dialogs.open('debug.' + id)
            self.window.debugger.active[id] = True
            self.window.debugger.update(True)

        # update menu
        self.update_menu()

    def update_menu(self):
        """Updates debug menu"""
        for id in self.window.debugger.ids:
            if id in self.window.debugger.active and self.window.debugger.active[id]:
                self.window.menu['debug.' + id].setChecked(True)
            else:
                self.window.menu['debug.' + id].setChecked(False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import datetime
import os
import requests
import openai


class Image:
    DIRNAME = "img"

    def __init__(self, config):
        """
        DALL-E Wrapper

        :param config: Config object
        """
        self.config = config
        if not self.config.initialized:
            self.config.init()

    def init(self):
        """Initialize OpenAI API key"""
        openai.api_key = self.config.data["api_key"]
        openai.organization = self.config.data["organization_key"]

    def generate(self, prompt, num=None):
        """
        Calls DALL-E API

        :param prompt: Prompt
        :param num: Number of variants
        :return: Images paths list
        """
        if num is None:
            num = int(input("How many variants generate? [1] ") or 1)
        print("Generating from: '{}'...".format(prompt))
        response = openai.Image.create(
            prompt=prompt,
            n=num,
            size=self.config.data["img_resolution"],
        )

        # generate and download images
        paths = []
        for i in range(num):
            url = response['data'][i]['url']
            res = requests.get(url)
            name = self.make_safe_filename(prompt) + "-" + datetime.date.today().strftime(
                "%Y-%m-%d") + "_" + datetime.datetime.now().strftime("%H-%M-%S") + "-" + str(i + 1) + ".png"
            path = os.path.join(self.config.path, self.DIRNAME, name)
            print("Downloading... [" + str(i + 1) + " of " + str(num) + "] to: " + path)
            open(path, "wb").write(res.content)
            paths.append(path)

        return paths

    def make_safe_filename(self, name):
        """
        Makes safe filename

        :param name: Filename to make safe
        :return: Safe filename
        """

        def safe_char(c):
            if c.isalnum():
                return c
            else:
                return "_"

        return "".join(safe_char(c) for c in name).rstrip("_")[:30]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import datetime
import os


class History:
    DIRNAME = "history"

    def __init__(self, config):
        """
        History handler

        :param config: Config
        """
        self.config = config

        if not self.config.initialized:
            self.config.init()

        self.path = os.path.join(self.config.path, self.DIRNAME)

    def save(self, text):
        """
        Saves text to history file

        :param text: Text
        """
        name = datetime.date.today().strftime("%Y_%m_%d") + ".txt"
        if not os.path.exists(self.path):
            try:
                os.makedirs(self.path)
            except Exception as e:
                print("Error creating history directory: " + str(e))
        if os.path.exists(self.path):
            f = os.path.join(self.path, name)
            try:
                with open(f, 'a', encoding="utf-8") as file:
                    prefix = ""
                    if self.config.data['store_history_time']:
                        prefix = datetime.datetime.now().strftime("%H:%M:%S") + ": "
                    file.write(prefix + text + "\n")
                    file.close()
            except Exception as e:
                print("Error saving history: " + str(e))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.12 08:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction

from ..utils import trans


class Theme:
    def __init__(self, window=None):
        """
        Theme controller

        :param window: main window object
        """
        self.window = window

    def toggle(self, name):
        """
        Toggles theme

        :param name: theme name
        """
        self.window.config.data['theme'] = name
        self.window.config.save()
        self.apply()
        self.window.set_theme(name + '.xml')
        self.update()

    def apply(self):
        """Applies theme fixes"""
        self.window.setStyleSheet(self.get_style('line_edit'))
        self.window.menu['menu.lang'].setStyleSheet(self.get_style('menu'))
        self.window.menu['menu.theme'].setStyleSheet(self.get_style('theme'))
        self.window.data['output'].setStyleSheet(self.get_style('chat_output'))
        self.window.data['output.timestamp'].setStyleSheet(self.get_style('checkbox'))
        self.window.data['input.send_enter'].setStyleSheet(self.get_style('radio'))
        self.window.data['input.send_shift_enter'].setStyleSheet(self.get_style('radio'))
        self.window.data['input.send_clear'].setStyleSheet(self.get_style('checkbox'))

        # dialog: ctx rename
        self.window.dialog['ctx.rename'].input.setStyleSheet(self.get_style('line_edit'))

        # ai, user names
        self.window.data['preset.ai_name'].setStyleSheet(self.get_style('line_edit'))
        self.window.data['preset.user_name'].setStyleSheet(self.get_style('line_edit'))

        # dialog: settings
        self.window.config_option['use_context'].box.setStyleSheet(self.get_style('checkbox'))
        self.window.config_option['store_history'].box.setStyleSheet(self.get_style('checkbox'))
        self.window.config_option['store_history_time'].box.setStyleSheet(self.get_style('checkbox'))

        # dialog: preset editor
        self.window.config_option['preset.chat'].box.setStyleSheet(self.get_style('checkbox'))
        self.window.config_option['preset.completion'].box.setStyleSheet(self.get_style('checkbox'))
        self.window.config_option['preset.img'].box.setStyleSheet(self.get_style('checkbox'))

    def get_style(self, element):
        """
        Returns style for element

        :param element: element name
        :return: style
        """
        theme = self.window.config.data['theme']

        # initial dark theme colors
        label = "fff"  # dark_teal
        chat_color = "fff"  # dark_teal
        checkbox_checked = "1de9b6"  # dark_teal
        checkbox_unchecked = "3a3f45"  # dark_teal

        # inverse colors for light themes
        if theme.startswith('light_'):
            label = "000"
            chat_color = "000"
            checkbox_unchecked = "f5f5f5"

        # Windows-OS invisible checkboxes fixes
        if theme.endswith('amber'):
            checkbox_checked = "ffd740"
        elif theme.endswith('blue'):
            checkbox_checked = "448aff"
        elif theme.endswith('cyan'):
            checkbox_checked = "4dd0e1"
        elif theme.endswith('lightgreen'):
            checkbox_checked = "8bc34a"
        elif theme.endswith('pink'):
            checkbox_checked = "ff4081"
        elif theme.endswith('purple'):
            checkbox_checked = "ab47bc"
        elif theme.endswith('red'):
            checkbox_checked = "ff1744"
        elif theme.endswith('teal'):
            checkbox_checked = "1de9b6"
        elif theme.endswith('yellow'):
            checkbox_checked = "ffff00"

        # get theme element style
        if element == 'checkbox':
            return "QCheckBox::indicator:checked { background-color: #" + checkbox_checked + "; } QCheckBox::indicator:unchecked { background-color: #" + checkbox_unchecked + "; }"  # Windows fix
        elif element == 'radio':
            return "QRadioButton::indicator:checked { background-color: #" + checkbox_checked + "; } QRadioButton::indicator:unchecked { background-color: #" + checkbox_unchecked + "; }"  # Windows fix
        elif element == 'menu':
            return "QMenu::indicator:checked { background-color: #" + checkbox_checked + "; } QMenu::indicator:unchecked { background-color: #" + checkbox_unchecked + "; }"  # Windows fix
        elif element == "tree_view":
            return "QTreeView { padding: 0px; margin: 0px; }; QTreeView::item { padding: 0px; margin: 0px; }"
        elif element == "chat_output":
            return 'color: #{}; font-size: {}px;'.format(chat_color, self.window.config.data['font_size'])
        elif element == "line_edit":
            return "QLineEdit { color: #" + label + "; }"
        elif element == "text_bold":
            return "font-weight: bold;"

    def update(self):
        """Updates theme menu"""
        for theme in self.window.menu['theme']:
            self.window.menu['theme'][theme].setChecked(False)
        current = self.window.config.data['theme']
        if current in self.window.menu['theme']:
            self.window.menu['theme'][current].setChecked(True)

    def get_themes_list(self):
        """
        Returns list of themes

        :return: list of themes
        """
        return ['dark_amber',
                'dark_blue',
                'dark_cyan',
                'dark_lightgreen',
                'dark_pink',
                'dark_purple',
                'dark_red',
                'dark_teal',
                'dark_yellow',
                'light_amber',
                'light_blue',
                'light_cyan',
                'light_cyan_500',
                'light_lightgreen',
                'light_pink',
                'light_purple',
                'light_red',
                'light_teal',
                'light_yellow']

    def trans_theme(self, theme):
        """
        Translates theme name

        :param theme: theme name
        :return: translated theme name
        """
        return theme.replace('_', ' ').title().replace('Dark ', trans('theme.dark') + ': ').replace('Light ', trans(
            'theme.light') + ': ')

    def setup(self):
        """Setups theme"""
        # setup menu
        themes = self.get_themes_list()
        for theme in themes:
            name = self.trans_theme(theme)
            self.window.menu['theme'][theme] = QAction(name, self.window, checkable=True)
            self.window.menu['theme'][theme].triggered.connect(
                lambda checked=None, theme=theme: self.window.controller.theme.toggle(theme))
            self.window.menu['menu.theme'].addAction(self.window.menu['theme'][theme])

        # apply theme
        theme = self.window.config.data['theme']
        self.toggle(theme)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

class ContextDebug:
    def __init__(self, window=None):
        """
        Context debug

        :param window: main window object
        """
        self.window = window
        self.id = 'context'

    def update(self):
        """Updates debug window."""
        self.window.debugger.begin(self.id)

        self.window.debugger.add(self.id, 'current_ctx', str(self.window.gpt.context.current_ctx))
        self.window.debugger.add(self.id, 'len(contexts)', str(len(self.window.gpt.context.contexts)))
        self.window.debugger.add(self.id, 'len(items)', str(len(self.window.gpt.context.items)))

        i = 0
        for item in self.window.gpt.context.items:
            prefix = '[{}] '.format(i)
            self.window.debugger.add(self.id, prefix + 'mode', str(item.mode))
            self.window.debugger.add(self.id, prefix + 'input_name', str(item.input_name))
            self.window.debugger.add(self.id, prefix + 'input', str(item.input))
            self.window.debugger.add(self.id, prefix + 'output_name', str(item.output_name))
            self.window.debugger.add(self.id, prefix + 'output', str(item.output))
            self.window.debugger.add(self.id, prefix + 'input_tokens',
                                     str(item.input_tokens))
            self.window.debugger.add(self.id, prefix + 'output_tokens',
                                     str(item.output_tokens))
            self.window.debugger.add(self.id, prefix + 'total_tokens', str(item.total_tokens))
            self.window.debugger.add(self.id, '------', '')
            i += 1

        self.window.debugger.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import openai

from .tokens import num_tokens_prompt, num_tokens_extra
from .context import Context, ContextItem
from .history import History


class Gpt:
    def __init__(self, config):
        """
        GPT Wrapper

        :param config: Config object
        """
        self.config = config
        self.context = Context(config)
        self.history = History(config)

        if not self.config.initialized:
            self.config.init()

    def init(self):
        """Initializes OpenAI API key"""
        openai.api_key = self.config.data["api_key"]
        openai.organization = self.config.data["organization_key"]

    def completion(self, prompt, max_tokens):
        """
        Calls OpenAI API for completion

        :param prompt: Prompt (user message)
        :param max_tokens: Max output tokens
        :return: Response dict
        """
        # build prompt message
        message = self.build_completion(prompt)

        # prepare stop word if user_name is set
        stop = None
        if self.config.data['user_name'] is not None and self.config.data['user_name'] != '':
            stop = [self.config.data['user_name'] + ':']

        response = openai.Completion.create(
            prompt=message,
            model=self.config.data['model'],
            max_tokens=int(max_tokens),
            temperature=self.config.data['temperature'],
            top_p=self.config.data['top_p'],
            frequency_penalty=self.config.data['frequency_penalty'],
            presence_penalty=self.config.data['presence_penalty'],
            stop=stop,
        )
        return response

    def chat(self, prompt, max_tokens):
        """
        Call OpenAI API for chat

        :param prompt: Prompt (user message)
        :param max_tokens: Max output tokens
        :return: Response dict
        """
        # build chat messages
        messages = self.build_chat_messages(prompt)
        response = openai.ChatCompletion.create(
            messages=messages,
            model=self.config.data['model'],
            max_tokens=int(max_tokens),
            temperature=self.config.data['temperature'],
            top_p=self.config.data['top_p'],
            frequency_penalty=self.config.data['frequency_penalty'],
            presence_penalty=self.config.data['presence_penalty'],
            stop=None,
        )
        return response

    def build_chat_messages(self, prompt):
        """
        Builds chat messages dict

        :param prompt: Prompt
        :return: Messages dict
        """
        messages = []

        model = self.config.data['model']
        used_tokens = self.count_used_tokens(prompt)
        max_tokens = self.config.data['max_total_tokens']

        # append initial (system) message
        if self.config.data['prompt'] is not None and self.config.data['prompt'] != "":
            messages.append({"role": "system", "content": self.config.data['prompt']})

        # append messages from context (memory)
        if self.config.data['use_context']:
            items = self.context.get_prompt_items(model, used_tokens, max_tokens)
            for item in items:
                # input
                if item.input_name is not None and item.input_name != "":
                    messages.append({"role": "user", "content": item.input})
                    # TODO: messages.append({"role": "user", "name": item.input_name, "content": item.input})
                else:
                    messages.append({"role": "user", "content": item.input})

                # output
                if item.output_name is not None and item.output_name != "":
                    messages.append({"role": "assistant", "content": item.output})
                    # TODO: messages.append({"role": "assistant", "name": item.output_name, "content": item.output})
                else:
                    messages.append({"role": "assistant", "content": item.output})

        # append current prompt
        if self.config.data['user_name'] is not None and self.config.data['user_name'] != "":
            messages.append({"role": "user", "content": str(prompt)})
            # TODO: messages.append({"role": "user", "name": self.config.data['user_name'], "content": str(prompt)})
        else:
            messages.append({"role": "user", "content": str(prompt)})

        return messages

    def build_completion(self, prompt):
        """
        Builds completion string

        :param prompt: Prompt (current)
        :return: Message string (parsed with context)
        """
        message = ""
        model = self.config.data['model']
        used_tokens = self.count_used_tokens(prompt)
        max_tokens = self.config.data['max_total_tokens']

        if self.config.data['prompt'] is not None and self.config.data['prompt'] != "":
            message += self.config.data['prompt']

        if self.config.data['use_context']:
            items = self.context.get_prompt_items(model, used_tokens, max_tokens)
            for item in items:
                if item.input_name is not None \
                        and item.output_name is not None \
                        and item.input_name != "" \
                        and item.output_name != "":
                    message += "\n" + item.input_name + ": " + item.input
                    message += "\n" + item.output_name + ": " + item.output
                else:
                    message += "\n" + item.input
                    message += "\n" + item.output

        if self.config.data['user_name'] is not None \
                and self.config.data['ai_name'] is not None \
                and self.config.data['user_name'] != "" \
                and self.config.data['ai_name'] != "":
            message += "\n" + self.config.data['user_name'] + ": " + str(prompt)
            message += "\n" + self.config.data['ai_name'] + ":"
        else:
            message += "\n" + str(prompt)

        return message

    def count_used_tokens(self, input_text):
        """
        Counts used tokens

        :param input_text: Input text
        :return: Used tokens
        """
        model = self.config.data['model']
        tokens = 0
        tokens += num_tokens_prompt(self.config.data['prompt'], self.config.data['user_name'],
                                    model)  # init (system) prompt
        tokens += num_tokens_prompt(input_text, self.config.data['user_name'], model)  # current input
        tokens += self.config.data['context_threshold']  # context threshold (reserved for next output)
        tokens += num_tokens_extra(model)  # extra tokens (required for output)
        return tokens

    def call(self, prompt, ctx=None):
        """
        Calls OpenAI API

        :param prompt: User input (prompt)
        :param ctx: Context item (memory)
        :return: Context item (memory)
        """
        # prepare max tokens
        model_tokens = self.config.get_model_tokens(self.config.data['model'])
        max_tokens = self.config.data['max_output_tokens']
        if max_tokens > model_tokens:
            max_tokens = model_tokens

        # minimum 1 token is required
        if max_tokens < 1:
            max_tokens = 1

        # store history
        if self.config.data['store_history']:
            self.history.save(prompt)

        # get response
        response = None
        if self.config.data['mode'] == "completion":
            response = self.completion(prompt, max_tokens)
        elif self.config.data['mode'] == "chat":
            response = self.chat(prompt, max_tokens)

        if response is None:
            return None

        # check for errors
        if "error" in response:
            print("Error: " + str(response["error"]))
            return None

        # get output
        output = ""
        if self.config.data['mode'] == "completion":
            output = response["choices"][0]["text"].strip()
        elif self.config.data['mode'] == "chat":
            output = response["choices"][0]["message"]["content"].strip()

        # store context (memory)
        if ctx is None:
            ctx = ContextItem(self.config.data['mode'])
            ctx.set_input(prompt, self.config.data['user_name'])

        ctx.set_output(output, self.config.data['ai_name'])
        ctx.set_tokens(response["usage"]["prompt_tokens"], response["usage"]["completion_tokens"])
        self.context.add(ctx)

        # store history
        if self.config.data['store_history']:
            self.history.save(output)

        return ctx

    def clear(self):
        """Clears context (memory)"""
        self.context.clear()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtGui import QTextCursor

from ..context import ContextItem
from ..utils import trans


class Input:
    def __init__(self, window=None):
        """
        Input controller

        :param window: main window
        """
        self.window = window

    def setup(self):
        """Setups input"""
        if self.window.config.data['send_clear']:
            self.window.data['input.send_clear'].setChecked(True)
        else:
            self.window.data['input.send_clear'].setChecked(False)

        if self.window.config.data['send_shift_enter']:
            self.window.data['input.send_shift_enter'].setChecked(True)
            self.window.data['input.send_enter'].setChecked(False)
        else:
            self.window.data['input.send_enter'].setChecked(True)
            self.window.data['input.send_shift_enter'].setChecked(False)

        self.window.data['input'].setFocus()

    def toggle_send_clear(self, value):
        """
        Toggles send clear

        :param value: value of the checkbox
        """
        self.window.config.data['send_clear'] = value

    def toggle_send_shift(self, value):
        """
        Toggles send with shift

        :param value: value of the checkbox
        """
        self.window.config.data['send_shift_enter'] = value

    def send_text(self, text):
        """
        Sends text to GPT

        :param text: text to send
        """
        self.window.set_status(trans('status.sending'))

        # create ctx item
        ctx = ContextItem()
        ctx.set_input(text, self.window.config.data['user_name'])
        self.window.controller.output.append_input(ctx)

        # call GPT
        try:
            ctx = self.window.gpt.call(text, ctx)
            self.window.controller.output.append_output(ctx)
            self.window.gpt.context.store()
            self.window.set_status(
                trans('status.tokens') + ": {} + {} = {}".format(ctx.input_tokens, ctx.output_tokens, ctx.total_tokens))
        except Exception as e:
            print(e)
            self.window.ui.dialogs.alert(str(e))
            self.window.set_status(trans('status.error'))

    def send(self):
        """
        Sends input text to API
        """
        text = self.window.data['input'].toPlainText().strip()
        if len(text) > 0:
            if self.window.config.data['send_clear']:
                self.window.data['input'].clear()

            # check API key
            if self.window.config.data['api_key'] is None or self.window.config.data['api_key'] == '':
                self.window.controller.launcher.show_api_monit()
                self.window.controller.ui.update()
                return

            # init api key if defined later
            self.window.gpt.init()
            self.window.images.init()

            # prepare context
            if len(self.window.gpt.context.contexts) == 0:
                self.window.gpt.context.new()
                self.window.controller.context.update()

            # send to API
            if self.window.config.data['mode'] == 'img':
                self.window.controller.image.send_text(text)
            else:
                self.window.controller.input.send_text(text)

        self.window.controller.ui.update()

    def append(self, text):
        """
        Appends text to input

        :param text: text to append
        """
        cur = self.window.data['input'].textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)
        s = str(text) + "\n"
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertBlock()
        self.window.data['input'].setTextCursor(cur)  # Update visible cursor



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.13 23:00:00                  #
# ================================================== #

from urllib.request import urlopen, Request
from packaging.version import parse as parse_version
import json
import ssl
from .utils import trans


class Updater:
    def __init__(self, window=None):
        """
        Updates handler

        :param window: main window
        """
        self.window = window

    def check(self):
        """Checks for updates"""
        print("Checking for updates...")
        url = self.window.website + "/api/version?v=" + str(self.window.version)
        try:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE

            req = Request(
                url=url,
                headers={'User-Agent': 'Mozilla/5.0'}
            )
            response = urlopen(req, context=ctx, timeout=3)
            data_json = json.loads(response.read())
            newest_version = data_json["version"]
            newest_build = data_json["build"]

            # changelog
            changelog = ""
            if "changelog" in data_json:
                changelog = data_json["changelog"]

            parsed_newest_version = parse_version(newest_version)
            parsed_current_version = parse_version(self.window.version)
            if parsed_newest_version > parsed_current_version:
                self.show_version_dialog(newest_version, newest_build, changelog)
            else:
                print("No updates available")
        except Exception as e:
            print("Failed to check for updates")
            print(e)

    def show_version_dialog(self, version, build, changelog):
        """
        Displays new version dialog

        :param version: version number
        :param build: build date
        :param changelog: changelog
        """
        txt = trans('update.new_version') + ": " + str(version) + " (" + trans('update.released') + ": " + str(
            build) + ")"
        txt += "\n" + trans('update.current_version') + ": " + self.window.version
        self.window.dialog['update'].changelog.setPlainText(changelog)
        self.window.dialog['update'].message.setText(txt)
        self.window.ui.dialogs.open('update')

    def patch(self):
        """Patch config files to current version"""
        try:
            self.patch_config()
            self.patch_models()
            self.patch_presets()
            # TODO: add context patcher
        except Exception as e:
            print("Failed to patch config files!")
            print(e)

    def patch_models(self):
        """Migrates models to current version"""
        data = self.window.config.models
        version = "0.0.0"
        updated = False
        if '__meta__' in data and 'version' in data['__meta__']:
            version = data['__meta__']['version']
        old = parse_version(version)
        current = parse_version(self.window.version)
        if old < current:
            if old < parse_version("0.9.1"):
                # apply meta only (not attached in 0.9.0)
                updated = True

        # update file
        if updated:
            print("Migrated models.json.")
            self.window.config.models = data
            self.window.config.save_models()

    def patch_presets(self):
        """Migrates presets to current version"""
        for k in self.window.config.presets:
            data = self.window.config.presets[k]
            version = "0.0.0"
            updated = False
            if '__meta__' in data and 'version' in data['__meta__']:
                version = data['__meta__']['version']
            old = parse_version(version)
            current = parse_version(self.window.version)
            if old < current:
                if old < parse_version("0.9.1"):
                    pass

            # update file
            if updated:
                print("Migrated presets.")
                self.window.config.presets[k] = data
                self.window.config.save_preset(k)

    def patch_config(self):
        """Migrates config to current version"""
        data = self.window.config.data
        version = "0.0.0"
        updated = False
        if '__meta__' in data and 'version' in data['__meta__']:
            version = data['__meta__']['version']
        old = parse_version(version)
        current = parse_version(self.window.version)
        if old < current:
            if old < parse_version("0.9.2"):
                print("Migrating config from 0.9.2...")
                keys_to_remove = ['ui.ctx.min_width',
                                  'ui.ctx.max_width',
                                  'ui.toolbox.min_width',
                                  'ui.toolbox.max_width',
                                  'ui.dialog.settings.width',
                                  'ui.dialog.settings.height',
                                  'ui.chatbox.font.color']
                for key in keys_to_remove:
                    if key in data:
                        del data[key]
                if 'theme' not in data:
                    data['theme'] = "dark_teal"
                updated = True

            if old < parse_version("0.9.1"):
                print("Migrating config from 0.9.1...")
                keys_to_remove = ['user_id', 'custom']  # not needed anymore
                for key in keys_to_remove:
                    if key in data:
                        del data[key]
                keys_to_add = ['organization_key']
                for key in keys_to_add:
                    if key not in data:
                        data[key] = ""
                updated = True

        # update file
        if updated:
            print("Migrated config.json.")
            self.window.config.data = data
            self.window.config.save_config()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from ..utils import trans


class Context:
    def __init__(self, window=None):
        """
        Context controller

        :param window: main window object
        """
        self.window = window

    def setup(self):
        """Setups context"""
        self.window.gpt.context.load_list()
        if len(self.window.gpt.context.contexts) == 0:
            self.new()
        else:
            ctx = self.window.config.data['ctx']
            if ctx is not None and ctx in self.window.gpt.context.contexts:
                self.window.gpt.context.current_ctx = ctx
            else:
                self.window.gpt.context.current_ctx = self.window.gpt.context.get_first_ctx()

        self.load(self.window.gpt.context.current_ctx)

    def new(self):
        """Creates new context"""
        self.window.gpt.context.new()
        self.update()
        self.window.controller.output.clear()

    def update(self):
        """Updates context list"""
        items = self.window.gpt.context.get_list()
        self.window.ui.contexts.update_list('ctx.contexts', items)
        self.select_ctx_by_current()
        self.window.controller.ui.update()
        self.window.debugger.update(True)
        self.window.debugger.update(True)

        ctx = self.window.gpt.context.current_ctx
        if ctx is not None:
            self.window.config.data['ctx'] = ctx
            self.window.config.save()

    def load(self, ctx):
        """
        Loads context

        :param ctx: context name (id)
        """
        self.window.gpt.context.select(ctx)
        self.window.controller.output.clear()
        self.window.controller.output.append_context()
        self.update()

    def refresh(self):
        """Refreshes context"""
        self.load(self.window.gpt.context.current_ctx)

    def selection_change(self):
        """
        Selects context on list change
        """
        # TODO: implement this
        # idx = self.window.data['ctx.contexts'].currentIndex().row()
        # self.select(idx)
        pass

    def select(self, idx):
        """
        Selects context

        :param idx: context index
        """
        ctx = self.window.gpt.context.get_name_by_idx(idx)
        self.window.gpt.context.current_ctx = ctx
        self.load(ctx)

    def delete(self, idx, force=False):
        """
        Deletes context

        :param idx: context index
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm('ctx_delete', idx, trans('ctx.delete.confirm'))
            return

        ctx = self.window.gpt.context.get_name_by_idx(idx)
        self.window.gpt.context.delete_ctx(ctx)
        if self.window.gpt.context.current_ctx == ctx:
            self.window.gpt.context.current_ctx = None
            self.window.controller.output.clear()
        self.update()

    def delete_history(self, force=False):
        """
        Deletes context history item

        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm('ctx_delete_all', '', trans('ctx.delete.all.confirm'))
            return

        self.window.gpt.context.delete_all_ctx()
        self.update()

    def rename(self, idx):
        """
        Renames context

        :param idx: context index
        """
        ctx = self.window.gpt.context.get_name_by_idx(idx)
        ctx_data = self.window.gpt.context.get_context_by_name(ctx)
        self.window.dialog['ctx.rename'].input.setText(ctx_data['name'])
        self.window.dialog['ctx.rename'].current = ctx
        self.window.dialog['ctx.rename'].show()
        self.update()

    def select_ctx_by_current(self):
        """Selects ctx by current"""
        ctx = self.window.gpt.context.current_ctx
        items = self.window.gpt.context.get_list()
        if ctx in items:
            idx = list(items.keys()).index(ctx)
            current = self.window.models['ctx.contexts'].index(idx, 0)
            self.window.data['ctx.contexts'].setCurrentIndex(current)

    def update_name(self, ctx, name):
        """
        Updates context name

        :param ctx: context name (id)
        :param name: context name
        """
        self.window.gpt.context.contexts[ctx]['name'] = name
        self.window.gpt.context.dump_context(ctx)
        self.window.dialog['ctx.rename'].close()
        self.update()

    def dismiss_rename(self):
        """Dismisses rename dialog"""
        self.window.dialog['ctx.rename'].close()

    def add(self, ctx):
        """
        Adds context

        :param ctx: context name (id)
        """
        self.window.gpt.context.add(ctx)
        self.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.12 08:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QScrollArea, QWidget

from ..widgets import SettingsInput, SettingsSlider, SettingsCheckbox, SettingsDialog
from ...utils import trans


class Settings:
    def __init__(self, window=None):
        """
        Settings dialog

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """Setups settings dialog"""

        id = "settings"
        path = self.window.config.path

        self.window.data['settings.btn.defaults'] = QPushButton(trans("dialog.settings.btn.defaults"))
        self.window.data['settings.btn.save'] = QPushButton(trans("dialog.settings.btn.save"))
        self.window.data['settings.btn.defaults'].clicked.connect(
            lambda: self.window.settings.load_default_settings())
        self.window.data['settings.btn.save'].clicked.connect(
            lambda: self.window.controller.settings.save(id))

        bottom_layout = QHBoxLayout()
        bottom_layout.addWidget(self.window.data['settings.btn.defaults'])
        bottom_layout.addWidget(self.window.data['settings.btn.save'])

        self.window.path_label[id] = QLabel(str(path))
        self.window.path_label[id].setStyleSheet("font-weight: bold;")
        self.window.config_option['api_key'] = SettingsInput(self.window, 'api_key')
        self.window.config_option['organization_key'] = SettingsInput(self.window, 'organization_key')
        self.window.config_option['img_resolution'] = SettingsInput(self.window, 'img_resolution')
        self.window.config_option['temperature'] = SettingsSlider(self.window, 'temperature',
                                                                  '', 0, 200,
                                                                  1, 100)
        self.window.config_option['top_p'] = SettingsSlider(self.window, 'top_p',
                                                            '', 0, 200,
                                                            1, 100)
        self.window.config_option['frequency_penalty'] = SettingsSlider(self.window, 'frequency_penalty',
                                                                        '', 0, 200,
                                                                        1, 0)
        self.window.config_option['presence_penalty'] = SettingsSlider(self.window, 'presence_penalty',
                                                                       '', 0, 200,
                                                                       1, 0)
        self.window.config_option['use_context'] = SettingsCheckbox(self.window, 'use_context',
                                                                    trans('settings.use_context'))
        self.window.config_option['store_history'] = SettingsCheckbox(self.window, 'store_history',
                                                                      trans('settings.store_history'))
        self.window.config_option['store_history_time'] = SettingsCheckbox(self.window, 'store_history_time',
                                                                           trans('settings.store_history_time'))
        self.window.config_option['context_threshold'] = SettingsSlider(self.window, 'context_threshold',
                                                                        '', 0, 1000,
                                                                        1, 80)
        self.window.config_option['max_output_tokens'] = SettingsSlider(self.window, 'max_output_tokens',
                                                                        '', 1, 32000,
                                                                        1, 50)
        self.window.config_option['max_total_tokens'] = SettingsSlider(self.window, 'max_total_tokens',
                                                                       '', 1, 32000,
                                                                       1, 400)
        self.window.config_option['font_size'] = SettingsSlider(self.window, 'font_size',
                                                                '', 8, 20,
                                                                1, 12)

        options = {}
        options['api_key'] = self.add_option('settings.api_key', self.window.config_option['api_key'], True)
        options['organization_key'] = self.add_option('settings.organization_key',
                                                      self.window.config_option['organization_key'], True)
        options['temperature'] = self.add_option('settings.temperature', self.window.config_option['temperature'])
        options['top_p'] = self.add_option('settings.top_p', self.window.config_option['top_p'])
        options['frequency_penalty'] = self.add_option('settings.frequency_penalty',
                                                       self.window.config_option['frequency_penalty'])
        options['presence_penalty'] = self.add_option('settings.presence_penalty',
                                                      self.window.config_option['presence_penalty'])
        options['use_context'] = self.add_raw_option(self.window.config_option['use_context'])
        options['store_history'] = self.add_raw_option(self.window.config_option['store_history'])
        options['store_history_time'] = self.add_raw_option(self.window.config_option['store_history_time'])
        options['context_threshold'] = self.add_option('settings.context_threshold',
                                                       self.window.config_option['context_threshold'])
        options['max_output_tokens'] = self.add_option('settings.max_output_tokens',
                                                       self.window.config_option['max_output_tokens'])
        options['max_total_tokens'] = self.add_option('settings.max_total_tokens',
                                                      self.window.config_option['max_total_tokens'])
        options['font_size'] = self.add_option('settings.font_size',
                                               self.window.config_option['font_size'])
        options['img_resolution'] = self.add_option('settings.img_resolution',
                                                    self.window.config_option['img_resolution'])

        rows = QVBoxLayout()
        rows.addLayout(options['api_key'])
        rows.addLayout(options['organization_key'])

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)

        scroll_content = QVBoxLayout()
        scroll_content.addLayout(options['use_context'])
        scroll_content.addLayout(options['store_history'])
        scroll_content.addLayout(options['store_history_time'])
        scroll_content.addLayout(options['font_size'])
        scroll_content.addLayout(options['context_threshold'])
        scroll_content.addLayout(options['max_output_tokens'])
        scroll_content.addLayout(options['max_total_tokens'])
        scroll_content.addLayout(options['temperature'])
        scroll_content.addLayout(options['top_p'])
        scroll_content.addLayout(options['frequency_penalty'])
        scroll_content.addLayout(options['presence_penalty'])
        scroll_content.addLayout(options['img_resolution'])

        scroll_widget = QWidget()
        scroll_widget.setLayout(scroll_content)

        scroll.setWidget(scroll_widget)

        layout = QVBoxLayout()
        layout.addLayout(rows)
        layout.addWidget(scroll)
        layout.addLayout(bottom_layout)

        self.window.dialog['config.' + id] = SettingsDialog(self.window, id)
        self.window.dialog['config.' + id].setLayout(layout)
        self.window.dialog['config.' + id].setWindowTitle(trans('dialog.settings'))

    def add_option(self, title, option, bold=False):
        """
        Adds option

        :param title: Title
        :param option: Option
        :param bold: Bold title
        """
        label_key = title + '.label'
        self.window.data[label_key] = QLabel(trans(title))
        if bold:
            self.window.data[label_key].setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        layout = QHBoxLayout()
        layout.addWidget(self.window.data[label_key])
        layout.addWidget(option)

        if title == 'settings.api_key':
            self.window.data[label_key].setMinimumHeight(60)
        return layout

    def add_raw_option(self, option):
        """
        Adds raw option row

        :param option: Option
        """
        layout = QHBoxLayout()
        layout.addWidget(option)
        return layout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QPushButton, QRadioButton, QCheckBox

from .status import Status
from .widgets import ChatInput
from ..utils import trans


class Input:
    def __init__(self, window=None):
        """
        Input UI

        :param window: main UI window object
        """
        self.window = window
        self.status = Status(window)

    def setup(self):
        """
        Setups input

        :return: QVBoxLayout
        """
        # input textarea
        self.window.data['input'] = ChatInput(self.window)

        # status
        status_layout = self.status.setup()
        status_layout.setAlignment(Qt.AlignLeft)

        # send options
        self.window.data['input.send_enter'] = QRadioButton(trans("input.radio.enter"))
        self.window.data['input.send_enter'].setStyleSheet(
            self.window.controller.theme.get_style('radio'))  # Windows fix
        self.window.data['input.send_enter'].clicked.connect(
            lambda: self.window.controller.input.toggle_send_shift(
                not self.window.data['input.send_enter'].isChecked()))
        self.window.data['input.send_shift_enter'] = QRadioButton(trans("input.radio.enter_shift"))
        self.window.data['input.send_shift_enter'].setStyleSheet(
            self.window.controller.theme.get_style('radio'))  # Windows fix
        self.window.data['input.send_shift_enter'].clicked.connect(
            lambda: self.window.controller.input.toggle_send_shift(
                self.window.data['input.send_shift_enter'].isChecked()))
        self.window.data['input.send_clear'] = QCheckBox(trans('input.send_clear'))
        self.window.data['input.send_clear'].setStyleSheet(
            self.window.controller.theme.get_style('checkbox'))  # Windows fix
        self.window.data['input.send_clear'].stateChanged.connect(
            lambda: self.window.controller.input.toggle_send_clear(self.window.data['input.send_clear'].isChecked()))

        # send button
        self.window.data['input.send_btn'] = QPushButton(trans("input.btn.send"))
        self.window.data['input.send_btn'].clicked.connect(
            lambda: self.window.controller.input.send())

        # send layout
        send_layout = QHBoxLayout()
        send_layout.addWidget(self.window.data['input.send_clear'])
        send_layout.addWidget(self.window.data['input.send_enter'])
        send_layout.addWidget(self.window.data['input.send_shift_enter'])
        send_layout.addWidget(self.window.data['input.send_btn'])
        send_layout.setAlignment(Qt.AlignRight)

        # bottom layout
        bottom_layout = QHBoxLayout()
        bottom_layout.addLayout(status_layout)
        bottom_layout.addLayout(send_layout)

        # header
        self.window.data['input.label'] = QLabel(trans("input.label"))
        self.window.data['input.label'].setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        self.window.data['input.counter'] = QLabel("")

        header = QHBoxLayout()
        header.addWidget(self.window.data['input.label'])
        header.addWidget(self.window.data['input.counter'], alignment=Qt.AlignRight)

        # input layout
        layout = QVBoxLayout()
        layout.addLayout(header)
        layout.addWidget(self.window.data['input'])
        layout.addLayout(bottom_layout)

        return layout



# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.11 05:00:00                  #
# ================================================== #

from .model import Model
from .presets import Presets
from .plugins import Plugins
from .debug import Debug
from .settings import Settings
from .info import Info
from .input import Input
from .output import Output
from .context import Context
from .confirm import Confirm
from .ui import UI
from .launcher import Launcher
from .lang import Lang
from .image import Image
from .theme import Theme


class Controller:
    def __init__(self, window=None):
        """
        Main controller

        :param window: main window object
        """
        self.window = window
        self.model = Model(window)
        self.presets = Presets(window)
        self.plugins = Plugins(window)
        self.debug = Debug(window)
        self.settings = Settings(window)
        self.info = Info(window)
        self.input = Input(window)
        self.output = Output(window)
        self.context = Context(window)
        self.confirm = Confirm(window)
        self.ui = UI(window)
        self.launcher = Launcher(window)
        self.lang = Lang(window)
        self.image = Image(window)
        self.theme = Theme(window)

    def setup(self):
        """Setups controller"""
        # init material theme
        self.theme.setup()

        # setup all controllers
        self.lang.setup()
        self.model.setup()
        self.input.setup()
        self.output.setup()
        self.context.setup()
        self.ui.setup()
        self.info.setup()
        self.launcher.setup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.04.13 23:00:00                  #
# ================================================== #

import json
import os
import configparser
import io

from .config import Config


class Locale:
    def __init__(self):
        """Locale handler"""
        self.config = Config()
        self.config.init(False)
        self.lang = self.config.data['lang']
        self.data = {}
        self.load(self.lang)

    def reload(self):
        """
        Reload translations
        """
        self.config.load_config()
        self.lang = self.config.data['lang']
        self.load(self.lang)

    def load(self, lang):
        """
        Loads translation ini file

        :param lang: language code
        """
        if type(lang) is not str:
            lang = 'en'
        path = os.path.join(self.config.get_root_path(), 'data', 'locale', 'locale.' + lang + '.ini')
        if not os.path.exists(path):
            print("FATAL ERROR: {} not found!".format(path))
            return None
        try:
            ini = configparser.ConfigParser()
            data = io.open(path, mode="r", encoding="utf-8")
            ini.read_string(data.read())
            self.data = dict(ini.items('LOCALE'))
        except Exception as e:
            print(e)

    def get(self, key):
        """
        Returns translation for key

        :param key: translation key
        :return: translated string
        """
        if key in self.data:
            return self.data[key].replace('\\n', "\n")
        else:
            return key



from setuptools import setup, find_packages

VERSION = '0.9.3'
DESCRIPTION = 'A GPT4, GPT3, ChatGPT and DALL-E 2 Desktop App with chatbot, text completion and image generation'
LONG_DESCRIPTION = 'A package containing a GPT4, GPT3, ChatGPT and DALL-E 2 desktop chatbot, ' \
                   'text completion and image generation app, using OpenAI API and your own API Key. ' \
                   'It includes context memory and history, editable presets, customizable UI and more.'

setup(
    name='pygpt-net',
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    license='MIT',
    author="Marcin Szczygliński",
    author_email='info@pygpt.net',
    maintainer="Marcin Szczygliński",
    maintainer_email='info@pygpt.net',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'': ['CHANGELOG.txt', 'data/*', 'data/locale/*', 'data/config/*', 'data/config/presets/*']},
    url='https://github.com/szczyglis-dev/py-gpt',
    keywords='py_gpt, py-gpt, pygpt, desktop, app, gpt, gpt4, gpt3, chatgpt, dall-e, chat, chatbot, text completion,'
             'image generation, ai, api, openai, api key, context memory, history, presets, ui, qt, pyside',
    install_requires=[
        'openai>=0.27.4',
        'packaging>=23.0',
        'pyinstaller>=5.9.0',
        'PySide6-Essentials>=6.4.2',
        'qt-material>=2.14',
        'show-in-file-manager>=1.1.4',
        'tiktoken>=0.3.3',
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import os
import re

from ..utils import trans


class Presets:
    def __init__(self, window=None):
        """
        Presets controller

        :param window: main window object
        """
        self.window = window

    def update_field(self, id, value, preset=None, current=False):
        """
        Updates preset field from editor

        :param id: field id
        :param value: field value
        :param preset: preset name (ID / filename)
        :param current: if True, updates current preset
        """
        if preset is not None and preset != "":
            if preset in self.window.config.presets:
                if id == 'preset.ai_name':
                    self.window.config.presets[preset]['ai_name'] = value
                elif id == 'preset.user_name':
                    self.window.config.presets[preset]['user_name'] = value
                elif id == 'preset.prompt':
                    self.window.config.presets[preset]['prompt'] = value
                elif id == 'preset.temperature' or id == 'current_temperature':
                    self.window.config.presets[preset]['temperature'] = float(value)

        # update current data
        if current:
            if id == 'preset.ai_name':
                self.window.config.data['ai_name'] = value
            elif id == 'preset.user_name':
                self.window.config.data['user_name'] = value
            elif id == 'preset.prompt':
                self.window.config.data['prompt'] = value
            elif id == 'preset.temperature' or id == 'current_temperature':
                self.window.config.data['temperature'] = float(value)

        self.window.controller.ui.update()

    def edit(self, idx=None):
        """
        Opens preset editor

        :param idx: preset index (row index)
        """
        preset = None
        if idx is not None:
            mode = self.window.config.data['mode']
            preset = self.window.config.get_preset_by_idx(idx, mode)

        self.init_editor(preset)
        self.window.ui.dialogs.open_editor('editor.preset.presets', idx)

    def init_editor(self, preset=None):
        """
        Initializes preset editor

        :param preset: preset name (ID / filename)
        """
        data = {}
        data['ai_name'] = ""
        data['user_name'] = ""
        data['prompt'] = ""
        data['temperature'] = 1.0
        data['img'] = False
        data['chat'] = False
        data['completion'] = False
        data['name'] = ""
        data['filename'] = ""

        if preset is not None and preset != "":
            if preset in self.window.config.presets:
                data = self.window.config.presets[preset]
                data['filename'] = preset

        if data['name'] is None:
            data['name'] = ""
        if data['ai_name'] is None:
            data['ai_name'] = ""
        if data['user_name'] is None:
            data['user_name'] = ""
        if data['prompt'] is None:
            data['prompt'] = ""
        if data['filename'] is None:
            data['filename'] = ""

        if preset is None:
            mode = self.window.config.data['mode']
            if mode == 'chat':
                data['chat'] = True
            elif mode == 'completion':
                data['completion'] = True
            elif mode == 'img':
                data['img'] = True

        self.window.controller.settings.change('preset.filename', data['filename'], 'preset.editor')
        self.window.controller.settings.change('preset.ai_name', data['ai_name'], 'preset.editor')
        self.window.controller.settings.change('preset.user_name', data['user_name'], 'preset.editor')
        self.window.controller.settings.change('preset.prompt', data['prompt'], 'preset.editor')
        self.window.controller.settings.change('preset.name', data['name'], 'preset.editor')
        self.window.controller.settings.apply('preset.temperature', data['temperature'], '', 'preset.editor')
        self.window.controller.settings.toggle('preset.img', data['img'], 'preset.editor')
        self.window.controller.settings.toggle('preset.chat', data['chat'], 'preset.editor')
        self.window.controller.settings.toggle('preset.completion', data['completion'], 'preset.editor')

    def save(self, force=False):
        """
        Saves preset

        :param force: force overwrite file
        """
        preset = self.window.config_option['preset.filename'].text()
        if preset is None or preset == "" or preset.startswith('current.'):
            self.window.ui.dialogs.alert(trans('alert.preset.empty_id'))
            self.window.set_status(trans('status.preset.empty_id'))
            return

        preset = self.window.controller.presets.validate_filename(preset)

        if preset not in self.window.config.presets:
            self.window.config.presets[preset] = {}

        filepath = os.path.join(self.window.config.path, 'presets', preset + '.json')

        # if exists then show confirmation dialog
        if os.path.exists(filepath) and not force:
            self.window.ui.dialogs.confirm('preset_exists', preset, trans('confirm.preset.overwrite'))
            return

        # check if at least one mode is enabled
        is_chat = self.window.config_option['preset.chat'].box.isChecked()
        is_completion = self.window.config_option['preset.completion'].box.isChecked()
        is_img = self.window.config_option['preset.img'].box.isChecked()

        # if not chat or completion then show warning
        if not is_chat and not is_completion and not is_img:
            self.window.ui.dialogs.alert(trans('alert.preset.no_chat_completion'))
            return

        # assign data from fields to preset
        self.assign_data(preset)

        # save file
        self.window.config.save_preset(preset)
        self.window.controller.model.update()

        self.window.ui.dialogs.close('editor.preset.presets')
        self.window.set_status(trans('status.preset.saved'))

    def assign_data(self, preset):
        """
        Assigns data from fields to preset

        :param preset: preset name (ID / filename)
        """
        name = self.window.config_option['preset.name'].text()
        if name is None or name == "":
            name = preset + " " + trans('preset.untitled')
        self.window.config.presets[preset]['name'] = name
        self.window.config.presets[preset]['ai_name'] = self.window.config_option['preset.ai_name'].text()
        self.window.config.presets[preset]['user_name'] = self.window.config_option['preset.user_name'].text()
        self.window.config.presets[preset]['prompt'] = self.window.config_option['preset.prompt'].toPlainText()
        self.window.config.presets[preset]['temperature'] = float(
            self.window.config_option['preset.temperature'].input.text())
        self.window.config.presets[preset]['img'] = self.window.config_option['preset.img'].box.isChecked()
        self.window.config.presets[preset]['chat'] = self.window.config_option['preset.chat'].box.isChecked()
        self.window.config.presets[preset]['completion'] = self.window.config_option[
            'preset.completion'].box.isChecked()

    def duplicate(self, idx=None):
        """
        Duplicates preset

        :param idx: preset index (row index)
        """
        if idx is not None:
            mode = self.window.config.data['mode']
            preset = self.window.config.get_preset_by_idx(idx, mode)
            if preset is not None and preset != "":
                if preset in self.window.config.presets:
                    new_id = self.window.config.duplicate_preset(preset)
                    self.window.config.data['preset'] = new_id
                    self.window.controller.model.update()
                    idx = self.window.config.get_preset_idx(mode, new_id)
                    self.edit(idx)
                    self.window.set_status(trans('status.preset.duplicated'))

    def clear(self, force=False):
        """
        Clears preset data

        :param force: force clear data
        """
        preset = self.window.config.data['preset']

        if not force:
            self.window.ui.dialogs.confirm('preset_clear', '', trans('confirm.preset.clear'))
            return

        self.window.config.data['prompt'] = ""
        self.window.config.data['ai_name'] = ""
        self.window.config.data['user_name'] = ""
        self.window.config.data['temperature'] = 1.0

        if preset is not None and preset != "":
            if preset in self.window.config.presets:
                self.window.config.presets[preset]['ai_name'] = ""
                self.window.config.presets[preset]['user_name'] = ""
                self.window.config.presets[preset]['prompt'] = ""
                self.window.config.presets[preset]['temperature'] = 1.0
                self.window.controller.model.update()

        self.window.set_status(trans('status.preset.cleared'))

    def delete(self, idx=None, force=False):
        """
        Deletes preset

        :param idx: preset index (row index)
        :param force: force delete without confirmation
        """
        if idx is not None:
            mode = self.window.config.data['mode']
            preset = self.window.config.get_preset_by_idx(idx, mode)
            if preset is not None and preset != "":
                if preset in self.window.config.presets:
                    # if exists then show confirmation dialog
                    if not force:
                        self.window.ui.dialogs.confirm('preset_delete', idx, trans('confirm.preset.delete'))
                        return

                    if preset == self.window.config.data['preset']:
                        self.window.config.data['preset'] = None
                    self.window.config.delete_preset(preset, True)
                    self.window.controller.model.update()
                    self.window.set_status(trans('status.preset.deleted'))

    def from_current(self):
        """Loads from current prompt"""
        self.window.controller.settings.change('preset.ai_name', self.window.config.data['ai_name'], 'preset.editor')
        self.window.controller.settings.change('preset.user_name', self.window.config.data['user_name'],
                                               'preset.editor')
        self.window.controller.settings.change('preset.prompt', self.window.config.data['prompt'], 'preset.editor')
        self.window.controller.settings.apply('preset.temperature', self.window.config.data['temperature'], '',
                                              'preset.editor')

    def use(self):
        """Copies preset prompt to input"""
        text = self.window.data['preset.prompt'].toPlainText()
        self.window.controller.input.append(text)

    def validate_filename(self, value):
        """
        Validates filename

        :param value: filename
        :return: validated filename
        """
        # strip not allowed characters
        return re.sub(r'[^\w\s-]', '', value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QVBoxLayout, QLabel, QPushButton, QWidget

from ..ui.widgets import ContextSelectMenu
from ..utils import trans


class Contexts:
    def __init__(self, window=None):
        """
        Contexts UI

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """
        Setups contexts list

        :return: QVBoxLayout
        """
        # contexts
        contexts = self.setup_contexts()
        self.window.models['ctx.contexts'] = self.create_model(self.window)
        self.window.data['ctx.contexts'].setModel(self.window.models['ctx.contexts'])
        self.window.data['ctx.contexts'].selectionModel().selectionChanged.connect(
            lambda: self.window.controller.context.selection_change())

        ctx_widget = QWidget()
        ctx_widget.setLayout(contexts)
        ctx_widget.setContentsMargins(0, 0, 0, 0)

        layout = QVBoxLayout()
        layout.addWidget(ctx_widget)
        layout.setContentsMargins(0, 0, 0, 0)

        return layout

    def setup_contexts(self):
        """
        Setups contexts list

        :return: QVBoxLayout
        """
        id = 'ctx.contexts'
        self.window.data['contexts.new'] = QPushButton(trans('context.new'))
        self.window.data['contexts.new'].clicked.connect(
            lambda: self.window.controller.context.new())

        self.window.data[id] = ContextSelectMenu(self.window, id)
        self.window.data['contexts.label'] = QLabel(trans("ctx.contexts.label"))
        self.window.data['contexts.label'].setStyleSheet(self.window.controller.theme.get_style('text_bold'))
        layout = QVBoxLayout()
        layout.addWidget(self.window.data['contexts.label'])
        layout.addWidget(self.window.data[id])
        layout.addWidget(self.window.data['contexts.new'])

        self.window.models[id] = self.create_model(self.window)
        self.window.data[id].setModel(self.window.models[id])
        return layout

    def create_model(self, parent):
        """
        Creates list model

        :param parent: parent widget
        :return: QStandardItemModel
        """
        model = QStandardItemModel(0, 1, parent)
        return model

    def update_list(self, id, data):
        """
        Updates list

        :param id: ID of the list
        :param data: Data to update
        """
        self.window.models[id].removeRows(0, self.window.models[id].rowCount())
        i = 0
        for n in data:
            if 'name' in data[n] and 'date' in data[n]:
                self.window.models[id].insertRow(i)
                name = data[n]['name'] + ' (' + data[n]['date'] + ')'
                self.window.models[id].setData(self.window.models[id].index(i, 0), name)
                i += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import os
import re
from .locale import Locale

locale = None


def trans(key, reload=False):
    """
    Returns translation

    :param key: translation key
    :param reload: force reload translations
    :return: translated string
    """
    global locale
    if locale is None:
        locale = Locale()
    if reload:
        locale.reload()
    return locale.get(key)


def get_init_value(key="__version__"):
    """
    Returns version

    :return: version string
    """
    path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '__init__.py'))
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(key),
                       open(path, "r", encoding="utf-8").read())
    return result.group(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

import tiktoken


def num_tokens_from_string(string, model="gpt-3.5-turbo"):
    """
    Returns number of tokens from string

    :param string: string
    :param model: model name
    :return: number of tokens
    """
    try:
        if model is not None:
            encoding = tiktoken.encoding_for_model(model)
        else:
            encoding = tiktoken.get_encoding("cl100k_base")
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")
    try:
        return len(encoding.encode(string))
    except Exception as e:
        print(e)
        return 0


def num_tokens_extra(model="gpt-3.5-turbo"):
    """
    Returns number of extra tokens

    :param model: model name
    :return: number of tokens
    """
    tokens = 5
    if model is not None and model.startswith("gpt-4"):
        tokens = 6
    elif model is not None and model.startswith("text-davinci"):
        tokens = 5
    return tokens


def num_tokens_prompt(text, input_name, model="gpt-3.5-turbo"):
    """
    Returns number of tokens from prompt

    :param text: prompt text
    :param input_name: input name
    :param model: model name
    :return: number of tokens
    """
    try:
        if model is not None:
            encoding = tiktoken.encoding_for_model(model)
        else:
            encoding = tiktoken.get_encoding("cl100k_base")
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 0

    # tags required tokens
    tokens_per_message = 4  # message follows <|start|>{role/name}\n{content}<|end|>\n
    tokens_per_name = -1

    if model is not None and model.startswith("gpt-4"):
        tokens_per_message = 3
        tokens_per_name = 1
    elif model is not None and model.startswith("text-davinci"):
        tokens_per_message = 1
        tokens_per_name = 0

    try:
        num_tokens += num_tokens_from_string(text, model)
    except Exception as e:
        print(e)

    if input_name is not None and input_name != "":
        num_tokens += tokens_per_message + tokens_per_name
    else:
        num_tokens += tokens_per_message

    return num_tokens


def num_tokens_from_context_item(item, model="gpt-3.5-turbo"):
    """
    Returns number of tokens from context item

    :param item: context item
    :param model: model name
    :return: number of tokens
    """
    try:
        if model is not None:
            encoding = tiktoken.encoding_for_model(model)
        else:
            encoding = tiktoken.get_encoding("cl100k_base")
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")

    # tags required tokens
    tokens_per_message = 4  # message follows <|start|>{role/name}\n{content}<|end|>\n
    tokens_per_name = -1
    multiply = 2
    if model is not None and model.startswith("gpt-4"):
        tokens_per_message = 3
        tokens_per_name = 1
    elif model is not None and model.startswith("text-davinci"):
        tokens_per_message = 1
        tokens_per_name = 0
        multiply = 1

    num_tokens = 0
    num_tokens += tokens_per_message * multiply  # input + output

    try:
        num_tokens += len(encoding.encode(item.input))
    except Exception as e:
        print(e)

    try:
        num_tokens += len(encoding.encode(item.output))
    except Exception as e:
        print(e)

    try:
        num_tokens += len(encoding.encode('user'))
    except Exception as e:
        print(e)

    try:
        num_tokens += len(encoding.encode('assistant'))
    except Exception as e:
        print(e)

    if item.input_name != "" and item.input_name is not None:
        num_tokens += tokens_per_name
        try:
            num_tokens += len(encoding.encode(item.input_name))
        except Exception as e:
            print(e)
    if item.output_name != "" and item.output_name is not None:
        num_tokens += tokens_per_name
        try:
            num_tokens += len(encoding.encode(item.output_name))
        except Exception as e:
            print(e)

    return num_tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Created Date: 2023.04.09 20:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon
from ..utils import trans


class Menu:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: main UI window object
        """
        self.window = window

    def setup(self):
        """Setups all menus"""
        self.window.menu = {}
        self.setup_file()
        self.setup_plugins()
        self.setup_audio()
        self.setup_config()
        self.setup_lang()
        self.setup_theme()
        self.setup_about()

        # debug menu
        if self.window.config.data['debug']:
            self.setup_debug()

    def setup_file(self):
        """Setups file menu"""
        self.window.menu['app.exit'] = QAction(QIcon.fromTheme("application-exit"), trans("menu.file.exit"),
                                               self.window, shortcut="Ctrl+Q", triggered=self.window.close)

        self.window.menu['app.clear_history'] = QAction(trans("menu.file_clear_history"),
                                                        self.window)

        self.window.menu['app.clear_history'].triggered.connect(
            lambda: self.window.controller.context.delete_history())

        self.window.menu['menu.app'] = self.window.menuBar().addMenu(trans("menu.file"))
        self.window.menu['menu.app'].addAction(self.window.menu['app.clear_history'])
        self.window.menu['menu.app'].addAction(self.window.menu['app.exit'])

    def setup_plugins(self):
        """Setups plugins menu"""
        self.window.menu['plugins.empty'] = QAction(QIcon.fromTheme("help-about"), trans("coming_soon"),
                                                    self.window)
        self.window.menu['menu.plugins'] = self.window.menuBar().addMenu(trans("menu.plugins"))
        self.window.menu['menu.plugins'].addAction(self.window.menu['plugins.empty'])

    def setup_audio(self):
        """Setups audio menu"""
        self.window.menu['audio.empty'] = QAction(QIcon.fromTheme("help-about"), trans("coming_soon"),
                                                  self.window)
        self.window.menu['menu.audio'] = self.window.menuBar().addMenu(trans("menu.audio"))
        self.window.menu['menu.audio'].addAction(self.window.menu['audio.empty'])

    def setup_config(self):
        """Setups config menu"""
        self.window.menu['config.settings'] = QAction(trans("menu.config.settings"),
                                                      self.window)
        self.window.menu['config.edit.config'] = QAction(trans("menu.config.edit.config"),
                                                         self.window)
        self.window.menu['config.edit.models'] = QAction(trans("menu.config.edit.models"),
                                                         self.window)
        self.window.menu['config.open_dir'] = QAction(trans("menu.config.open_dir"),
                                                      self.window)
        self.window.menu['config.save'] = QAction(QIcon.fromTheme("document-save"), trans("menu.config.save"),
                                                  self.window)

        self.window.menu['config.settings'].triggered.connect(
            lambda: self.window.controller.settings.toggle_settings('settings'))

        self.window.menu['config.edit.config'].triggered.connect(
            lambda: self.window.controller.settings.toggle_editor('config.json'))

        self.window.menu['config.edit.models'].triggered.connect(
            lambda: self.window.controller.settings.toggle_editor('models.json'))

        self.window.menu['config.open_dir'].triggered.connect(
            lambda: self.window.controller.settings.open_config_dir())

        self.window.menu['config.save'].triggered.connect(
            lambda: self.window.controller.settings.save_all())

        self.window.menu['menu.config'] = self.window.menuBar().addMenu(trans("menu.config"))
        self.window.menu['menu.config'].addAction(self.window.menu['config.settings'])
        self.window.menu['menu.config'].addAction(self.window.menu['config.edit.config'])
        self.window.menu['menu.config'].addAction(self.window.menu['config.edit.models'])
        self.window.menu['menu.config'].addAction(self.window.menu['config.open_dir'])
        self.window.menu['menu.config'].addAction(self.window.menu['config.save'])

    def setup_debug(self):
        """Setups debug menu"""
        self.window.menu['debug.config'] = QAction(trans("menu.debug.config"), self.window, checkable=True)
        self.window.menu['debug.context'] = QAction(trans("menu.debug.context"), self.window, checkable=True)
        self.window.menu['debug.presets'] = QAction(trans("menu.debug.presets"), self.window, checkable=True)
        self.window.menu['debug.models'] = QAction(trans("menu.debug.models"), self.window, checkable=True)

        self.window.menu['debug.config'].triggered.connect(
            lambda: self.window.controller.debug.toggle('config'))
        self.window.menu['debug.context'].triggered.connect(
            lambda: self.window.controller.debug.toggle('context'))
        self.window.menu['debug.presets'].triggered.connect(
            lambda: self.window.controller.debug.toggle('presets'))
        self.window.menu['debug.models'].triggered.connect(
            lambda: self.window.controller.debug.toggle('models'))

        self.window.menu['menu.debug'] = self.window.menuBar().addMenu(trans("menu.debug"))
        self.window.menu['menu.debug'].addAction(self.window.menu['debug.config'])
        self.window.menu['menu.debug'].addAction(self.window.menu['debug.context'])
        self.window.menu['menu.debug'].addAction(self.window.menu['debug.presets'])
        self.window.menu['menu.debug'].addAction(self.window.menu['debug.models'])

    def setup_lang(self):
        """Setups lang menu"""
        self.window.menu['lang'] = {}
        self.window.menu['menu.lang'] = self.window.menuBar().addMenu(trans("menu.lang"))
        self.window.menu['menu.lang'].setStyleSheet(self.window.controller.theme.get_style('menu'))  # Windows fix

    def setup_theme(self):
        """Setups theme menu"""
        self.window.menu['theme'] = {}
        self.window.menu['menu.theme'] = self.window.menuBar().addMenu(trans("menu.theme"))
        self.window.menu['menu.theme'].setStyleSheet(self.window.controller.theme.get_style('menu'))  # Windows fix

    def setup_about(self):
        """Setups about menu"""
        self.window.menu['info.about'] = QAction(QIcon.fromTheme("help-about"), trans("menu.info.about"),
                                                 self.window)
        self.window.menu['info.changelog'] = QAction(trans("menu.info.changelog"),
                                                     self.window)
        self.window.menu['info.website'] = QAction(trans("menu.info.website"),
                                                   self.window)
        self.window.menu['info.github'] = QAction(trans("menu.info.github"),
                                                  self.window)

        self.window.menu['info.about'].triggered.connect(
            lambda: self.window.controller.info.toggle('about'))
        self.window.menu['info.changelog'].triggered.connect(
            lambda: self.window.controller.info.toggle('changelog'))
        self.window.menu['info.website'].triggered.connect(
            lambda: self.window.controller.info.goto_website())
        self.window.menu['info.github'].triggered.connect(
            lambda: self.window.controller.info.goto_github())

        self.window.menu['menu.about'] = self.window.menuBar().addMenu(trans("menu.info"))
        self.window.menu['menu.about'].addAction(self.window.menu['info.about'])
        self.window.menu['menu.about'].addAction(self.window.menu['info.changelog'])
        self.window.menu['menu.about'].addAction(self.window.menu['info.website'])
        self.window.menu['menu.about'].addAction(self.window.menu['info.github'])
